# source https://www.tensorflow.org/text/tutorials/transformer

import numpy as np
import tensorflow as tf

# positional encoding
from pyaromatics.keras_tools.esoteric_layers import SurrogatedStep
from pyaromatics.keras_tools.esoteric_layers.layer_scaling import LayerScaling
from pyaromatics.keras_tools.esoteric_layers.repu_variants import RePU


def layer_normalization(config):
    ln = tf.keras.layers.LayerNormalization(epsilon=1e-6)
    if 'nogradlayerscaling' in config:
        ln = LayerScaling(grad_through_maxmin=False)
    elif 'layerscaling' in config:
        ln = LayerScaling()
    elif 'layerscalingmeanstd' in config:
        ln = LayerScaling(center='mean', scale='std')
    elif 'layerscalingmeandist' in config:
        ln = LayerScaling(center='mean', scale='maxmin')

    return ln


def get_angles(pos, i, d_model):
    angle_rates = 1 / np.power(10000, (2 * (i // 2)) / np.float32(d_model))
    return pos * angle_rates


def positional_encoding(position, d_model):
    angle_rads = get_angles(np.arange(position)[:, np.newaxis],
                            np.arange(d_model)[np.newaxis, :],
                            d_model)

    # apply sin to even indices in the array; 2i
    angle_rads[:, 0::2] = np.sin(angle_rads[:, 0::2])

    # apply cos to odd indices in the array; 2i+1
    angle_rads[:, 1::2] = np.cos(angle_rads[:, 1::2])

    pos_encoding = angle_rads[np.newaxis, ...]

    return tf.cast(pos_encoding, dtype=tf.float32)


# padding and masks

def create_padding_mask(seq, pad_idx=0):
    seq = tf.cast(tf.math.equal(seq, pad_idx), tf.float32)

    # add extra dimensions to add the padding
    # to the attention logits.
    return seq[:, tf.newaxis, tf.newaxis, :]  # (batch_size, 1, 1, seq_len)


def create_look_ahead_mask(size):
    mask = 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0)
    return mask  # (seq_len, seq_len)


def create_masks(inp, tar, pad_idx):
    # Encoder padding mask
    enc_padding_mask = create_padding_mask(inp, pad_idx)

    # Used in the 2nd attention block in the decoder.
    # This padding mask is used to mask the encoder outputs.
    dec_padding_mask = create_padding_mask(inp, pad_idx)

    # Used in the 1st attention block in the decoder.
    # It is used to pad and mask future tokens in the input received by
    # the decoder.
    look_ahead_mask = create_look_ahead_mask(tf.shape(tar)[1])
    dec_target_padding_mask = create_padding_mask(tar)
    look_ahead_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)

    return enc_padding_mask, look_ahead_mask, dec_padding_mask


class PaddingLookAheadMasks(tf.keras.layers.Layer):
    def __init__(self, mask_type='look_ahead'):
        super().__init__()
        self.mask_type = mask_type

    def call(self, inputs, *args, **kwargs):

        ml, mp = 0., 0.
        if 'look_ahead' in self.mask_type:
            ml = create_look_ahead_mask(tf.shape(inputs)[1])

        if 'padding' in self.mask_type:
            mp = create_padding_mask(inputs)

        m = tf.maximum(ml, mp)

        return m


# attention and ff

def scaled_dot_product_attention_original(q, k, v, mask=None):
    """Calculate the attention weights.
    q, k, v must have matching leading dimensions.
    k, v must have matching penultimate dimension, i.e.: seq_len_k = seq_len_v.
    The mask has different shapes depending on its type(padding or look ahead)
    but it must be broadcastable for addition.

    Args:
      q: query shape == (..., seq_len_q, depth)
      k: key shape == (..., seq_len_k, depth)
      v: value shape == (..., seq_len_v, depth_v)
      mask: Float tensor with shape broadcastable
            to (..., seq_len_q, seq_len_k). Defaults to None.

    Returns:
      output, attention_weights
    """

    matmul_qk = tf.matmul(q, k, transpose_b=True)  # (..., seq_len_q, seq_len_k)

    # scale matmul_qk
    dk = tf.cast(tf.shape(k)[-1], tf.float32)
    scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)

    # add the mask to the scaled tensor.
    if mask is not None:
        scaled_attention_logits += (mask * -1e9)

    # softmax is normalized on the last axis (seq_len_k) so that the scores
    # add up to 1.
    attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)  # (..., seq_len_q, seq_len_k)

    output = tf.matmul(attention_weights, v)  # (..., seq_len_q, depth_v)

    return output, attention_weights


def scaled_dot_product_attention_spiking(q, k, v, mask):
    """Calculate the attention weights.
    q, k, v must have matching leading dimensions.
    k, v must have matching penultimate dimension, i.e.: seq_len_k = seq_len_v.
    The mask has different shapes depending on its type(padding or look ahead)
    but it must be broadcastable for addition.

    Args:
      q: query shape == (..., seq_len_q, depth)
      k: key shape == (..., seq_len_k, depth)
      v: value shape == (..., seq_len_v, depth_v)
      mask: Float tensor with shape broadcastable
            to (..., seq_len_q, seq_len_k). Defaults to None.

    Returns:
      output, attention_weights
    """

    matmul_qk = tf.matmul(q, k, transpose_b=True)  # (..., seq_len_q, seq_len_k)
    # scale matmul_qk
    dk = tf.cast(tf.shape(k)[-1], tf.float32)
    L = tf.cast(tf.shape(matmul_qk)[-1], tf.float32)
    # scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)
    scaled_attention_logits = matmul_qk / L / dk

    # add the mask to the scaled tensor.
    if mask is not None:
        scaled_attention_logits += (mask * -1e9)

    # softmax is normalized on the last axis (seq_len_k) so that the scores
    # add up to 1.
    # attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)  # (..., seq_len_q, seq_len_k)
    attention_weights = scaled_attention_logits  # (..., seq_len_q, seq_len_k)
    output = tf.matmul(attention_weights, v)  # (..., seq_len_q, depth_v)

    return output, attention_weights


def choose_attention(attention_type):
    if 'spiking_attention' in attention_type:

        surrogated_step_I = SurrogatedStep(attention_type)
        surrogated_step_II = SurrogatedStep(attention_type)
        scaled_dot_product_attention = lambda q, k, v, mask: scaled_dot_product_attention_spiking(
            surrogated_step_I(q), surrogated_step_II(k), v, mask)
    else:
        scaled_dot_product_attention = scaled_dot_product_attention_original

    return scaled_dot_product_attention


class MultiHeadAttention(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, config='original_transformer'):
        super(MultiHeadAttention, self).__init__()
        self.num_heads = num_heads
        self.d_model = d_model

        assert d_model % self.num_heads == 0

        self.depth = d_model // self.num_heads

        self.wq = tf.keras.layers.Dense(d_model, use_bias=False)
        self.wk = tf.keras.layers.Dense(d_model, use_bias=False)
        self.wv = tf.keras.layers.Dense(d_model, use_bias=False)

        self.dense = tf.keras.layers.Dense(d_model, use_bias=False)

        self.scaled_dot_product_attention = choose_attention(config)

    def split_heads(self, x, batch_size):
        """Split the last dimension into (num_heads, depth).
        Transpose the result such that the shape is (batch_size, num_heads, seq_len, depth)
        """
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, v, k, q, mask):
        batch_size = tf.shape(q)[0]

        q = self.wq(q)  # (batch_size, seq_len, d_model)
        k = self.wk(k)  # (batch_size, seq_len, d_model)
        v = self.wv(v)  # (batch_size, seq_len, d_model)

        q = self.split_heads(q, batch_size)  # (batch_size, num_heads, seq_len_q, depth)
        k = self.split_heads(k, batch_size)  # (batch_size, num_heads, seq_len_k, depth)
        v = self.split_heads(v, batch_size)  # (batch_size, num_heads, seq_len_v, depth)

        # scaled_attention.shape == (batch_size, num_heads, seq_len_q, depth)
        # attention_weights.shape == (batch_size, num_heads, seq_len_q, seq_len_k)
        scaled_attention, attention_weights = self.scaled_dot_product_attention(q, k, v, mask)

        # (batch_size, seq_len_q, num_heads, depth)
        scaled_attention = tf.transpose(scaled_attention, perm=[0, 2, 1, 3])

        # (batch_size, seq_len_q, d_model)
        concat_attention = tf.reshape(scaled_attention, (batch_size, -1, self.d_model))

        # (batch_size, seq_len_q, d_model)
        output = self.dense(concat_attention)

        return output, attention_weights


def point_wise_feed_forward_network(d_model, dff, config):
    activation = tf.keras.layers.ReLU()
    # kernel_initializer = 'he_normal'
    kernel_initializer = tf.keras.initializers.VarianceScaling(scale=2., mode='fan_in',
                                                               distribution='untruncated_normal')
    bias_initializer = 'zeros'

    # 2/(n_in + n_out) Glorot
    # 1/(n_in)

    if 'trainable_repu' in config:
        activation = RePU(trainable_p=True)
    elif 'pswish' in config:
        activation = RePU(base_activation='swish', trainable_p=True)
    elif 'swish' in config:
        activation = RePU(base_activation='swish')
        if 'critical' in config:
            kernel_initializer = tf.keras.initializers.VarianceScaling(scale=1.988, mode='fan_in',
                                                                       distribution='untruncated_normal')
            bias_initializer = tf.keras.initializers.RandomNormal(mean=0.0, stddev=tf.sqrt(0.555))
    elif 'guderman' in config:
        activation = RePU(base_activation='guderman')
        if 'critical' in config:
            kernel_initializer = tf.keras.initializers.VarianceScaling(scale=1.990, mode='fan_in',
                                                                       distribution='untruncated_normal')
            bias_initializer = tf.keras.initializers.RandomNormal(mean=0.0, stddev=tf.sqrt(0.103))

    return tf.keras.Sequential([
        tf.keras.layers.Dense(dff, kernel_initializer='glorot_normal', bias_initializer='zeros'),
        # (batch_size, seq_len, dff)
        activation,
        tf.keras.layers.Dense(d_model, kernel_initializer=kernel_initializer, bias_initializer=bias_initializer)
        # (batch_size, seq_len, d_model)
    ])


# encoder and decoder

class EncoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, rate=0.1, config='original_transformer', do_fn=None):
        super(EncoderLayer, self).__init__()
        self.mha = MultiHeadAttention(d_model, num_heads, config=config)
        self.ffn = point_wise_feed_forward_network(d_model, dff, config)

        self.layernorm1 = layer_normalization(config)
        self.layernorm2 = layer_normalization(config)

        self.dropout1 = tf.keras.layers.Dropout(rate) if do_fn is None else do_fn
        self.dropout2 = tf.keras.layers.Dropout(rate) if do_fn is None else do_fn

    def call(self, x, mask):
        attn_output, _ = self.mha(x, x, x, mask)  # (batch_size, input_seq_len, d_model)
        attn_output = self.dropout1(attn_output)
        out1 = self.layernorm1(x + attn_output)  # (batch_size, input_seq_len, d_model)

        ffn_output = self.dropout2(self.ffn(out1))  # (batch_size, input_seq_len, d_model)
        out2 = self.layernorm2(out1 + ffn_output)  # (batch_size, input_seq_len, d_model)

        return out2


class DecoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, rate=0.1, config='original_transformer'):
        super(DecoderLayer, self).__init__()

        self.mha1 = MultiHeadAttention(d_model, num_heads, config=config)
        self.mha2 = MultiHeadAttention(d_model, num_heads, config=config)

        self.ffn = point_wise_feed_forward_network(d_model, dff, config)

        self.layernorm1 = layer_normalization(config)
        self.layernorm2 = layer_normalization(config)
        self.layernorm3 = layer_normalization(config)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)
        self.dropout3 = tf.keras.layers.Dropout(rate)

    def call(self, x, enc_output,
             look_ahead_mask, padding_mask):
        # enc_output.shape == (batch_size, input_seq_len, d_model)

        attn1, attn_weights_block1 = self.mha1(x, x, x, look_ahead_mask)  # (batch_size, target_seq_len, d_model)
        attn1 = self.dropout1(attn1)
        out1 = self.layernorm1(attn1 + x)

        attn2, attn_weights_block2 = self.mha2(
            enc_output, enc_output, out1, padding_mask)  # (batch_size, target_seq_len, d_model)
        attn2 = self.dropout2(attn2)
        out2 = self.layernorm2(attn2 + out1)  # (batch_size, target_seq_len, d_model)

        ffn_output = self.ffn(out2)  # (batch_size, target_seq_len, d_model)
        ffn_output = self.dropout3(ffn_output)
        out3 = self.layernorm3(ffn_output + out2)  # (batch_size, target_seq_len, d_model)

        return out3, attn_weights_block1, attn_weights_block2


class GPTBlock(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, rate=0.1, config='original_transformer'):
        super().__init__()

        self.mha1 = MultiHeadAttention(d_model, num_heads, config=config)

        self.ffn = point_wise_feed_forward_network(d_model, dff, config)

        self.layernorm1 = layer_normalization
        self.layernorm3 = layer_normalization

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout3 = tf.keras.layers.Dropout(rate)

    def call(self, x, look_ahead_mask):
        # enc_output.shape == (batch_size, input_seq_len, d_model)

        attn1, attn_weights_block1 = self.mha1(x, x, x, look_ahead_mask)  # (batch_size, target_seq_len, d_model)
        attn1 = self.dropout1(attn1)
        out1 = self.layernorm1(attn1 + x)

        ffn_output = self.ffn(out1)  # (batch_size, target_seq_len, d_model)
        ffn_output = self.dropout3(ffn_output)
        out3 = self.layernorm3(ffn_output + out1)  # (batch_size, target_seq_len, d_model)

        return out3


class TransformerEncoder(tf.keras.layers.Layer):
    def get_config(self):
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(self.init_args.items()))

    def __init__(self, num_layers, d_model, num_heads, dff, input_vocab_size,
                 maximum_position_encoding, rate=0.1, config='original_transformer', **kwargs):
        self.init_args = dict(num_layers=num_layers, d_model=d_model, num_heads=num_heads, dff=dff,
                              input_vocab_size=input_vocab_size, maximum_position_encoding=maximum_position_encoding,
                              rate=rate, config=config)
        super().__init__(**kwargs)

        self.d_model = d_model
        self.num_layers = num_layers

        self.embedding = tf.keras.layers.Embedding(input_vocab_size, d_model)
        self.pos_encoding = positional_encoding(maximum_position_encoding,
                                                self.d_model)

        self.enc_layers = [EncoderLayer(d_model, num_heads, dff, rate, config=config)
                           for _ in range(num_layers)]

        self.dropout = tf.keras.layers.Dropout(rate)

    def call(self, x, mask=None, *args, **kwargs):
        seq_len = tf.shape(x)[1]

        # adding embedding and position encoding.
        x = self.embedding(x)  # (batch_size, input_seq_len, d_model)
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]

        x = self.dropout(x)

        for i in range(self.num_layers):
            x = self.enc_layers[i](x, mask)

        return x  # (batch_size, input_seq_len, d_model)


class TransformerDecoder(tf.keras.layers.Layer):
    def get_config(self):
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(self.init_args.items()))

    def __init__(self, num_layers, d_model, num_heads, dff, target_vocab_size,
                 maximum_position_encoding, rate=0.1, config='original_transformer'):
        self.init_args = dict(num_layers=num_layers, d_model=d_model, num_heads=num_heads, dff=dff,
                              target_vocab_size=target_vocab_size, maximum_position_encoding=maximum_position_encoding,
                              rate=rate, config=config)
        super().__init__()

        self.d_model = d_model
        self.num_layers = num_layers

        self.embedding = tf.keras.layers.Embedding(target_vocab_size, d_model)
        self.pos_encoding = positional_encoding(maximum_position_encoding, d_model)

        self.dec_layers = [DecoderLayer(d_model, num_heads, dff, rate, config=config)
                           for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(rate)

    def call(self, x, enc_output=None,
             look_ahead_mask=None, padding_mask=None, output_type='', *args, **kwargs):
        seq_len = tf.shape(x)[1]
        attention_weights = {}

        x = self.embedding(x)  # (batch_size, target_seq_len, d_model)
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]

        x = self.dropout(x)

        for i in range(self.num_layers):
            x, block1, block2 = self.dec_layers[i](x, enc_output, look_ahead_mask, padding_mask)

            attention_weights[f'decoder_layer{i + 1}_block1'] = block1
            attention_weights[f'decoder_layer{i + 1}_block2'] = block2

        if output_type == 'embedding_projection':
            x = tf.matmul(x, self.embedding.embeddings, transpose_b=True)
        # x.shape == (batch_size, target_seq_len, d_model)
        return x, attention_weights


class GPT(tf.keras.layers.Layer):
    def get_config(self):
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(self.init_args.items()))

    def __init__(self, num_layers, d_model, num_heads, dff, target_vocab_size,
                 maximum_position_encoding, pad_idx, rate=0.1, config='original_transformer', **kwargs):
        super().__init__(**kwargs)

        self.init_args = dict(num_layers=num_layers, d_model=d_model, num_heads=num_heads, dff=dff,
                              target_vocab_size=target_vocab_size, maximum_position_encoding=maximum_position_encoding,
                              rate=rate, pad_idx=pad_idx, config=config)

        self.d_model = d_model
        self.num_layers = num_layers
        self.pad_idx = pad_idx

        self.embedding = tf.keras.layers.Embedding(target_vocab_size, d_model)
        self.pos_encoding = positional_encoding(maximum_position_encoding, d_model)

        self.dec_layers = [GPTBlock(d_model, num_heads, dff, rate, config=config)
                           for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(rate)

    def call(self, inputs, output_type='', *args, **kwargs):
        sentence = inputs
        seq_len = tf.shape(sentence)[1]

        look_ahead_mask = create_look_ahead_mask(tf.shape(sentence)[1])
        dec_target_padding_mask = create_padding_mask(sentence, pad_idx=self.pad_idx)
        decoder_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)

        x = self.embedding(sentence)  # (batch_size, target_seq_len, d_model)
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]

        x = self.dropout(x)

        for i in range(self.num_layers):
            x = self.dec_layers[i](x, decoder_mask)

        if output_type == 'embedding_projection':
            x = tf.matmul(x, self.embedding.embeddings, transpose_b=True)
        # x.shape == (batch_size, target_seq_len, d_model)
        return x


class Transformer(tf.keras.Model):
    def __init__(self, num_layers, d_model, num_heads, dff, input_vocab_size,
                 target_vocab_size, pe_input, pe_target, pad_idx, rate=0.1, config='original_transformer'):
        super().__init__()
        self.encoder = TransformerEncoder(num_layers, d_model, num_heads, dff,
                                          input_vocab_size, pe_input, rate, config=config)

        self.decoder = TransformerDecoder(num_layers, d_model, num_heads, dff,
                                          target_vocab_size, pe_target, rate, config=config)

        # self.final_layer = tf.keras.layers.Dense(target_vocab_size)
        self.pad_idx = pad_idx

    def call(self, inputs):
        # Keras models prefer if you pass all your inputs in the first argument
        inp, tar = inputs

        enc_padding_mask, look_ahead_mask, dec_padding_mask = create_masks(inp, tar, pad_idx=self.pad_idx)

        enc_output = self.encoder(inp, enc_padding_mask)  # (batch_size, inp_seq_len, d_model)

        # dec_output.shape == (batch_size, tar_seq_len, d_model)
        dec_output, attention_weights = self.decoder(
            tar, enc_output, look_ahead_mask, dec_padding_mask, output_type='embedding_projection')

        # final_output = self.final_layer(dec_output)  # (batch_size, tar_seq_len, target_vocab_size)

        return dec_output, attention_weights


def test_gpt():
    from pyaromatics.LeanguageTreatmentTools.random_language import random_indices
    vocab_size = 20
    pad_idx = 3
    src_tokens = random_indices(vocab_size, pad_idx=pad_idx)
    gpt_layer = GPT(num_layers=2, d_model=2, num_heads=2, dff=2, target_vocab_size=vocab_size,
                    maximum_position_encoding=vocab_size, pad_idx=pad_idx, rate=0.1)

    output = gpt_layer(src_tokens, output_type='embedding_projection')
    print(output.shape)


if __name__ == '__main__':
    # test_gpt()
    maxlen = 128
    vocab_size = 32000
    transformer = Transformer(num_layers=6, d_model=512, num_heads=8, dff=2048, input_vocab_size=vocab_size,
                              target_vocab_size=vocab_size, pe_input=maxlen, pe_target=maxlen, pad_idx=0)
    batch = np.random.choice(vocab_size, size=(3, maxlen))
    out = transformer((batch, batch))
    transformer.summary()
    print(transformer.count_params())
    print(out[0].shape)
